/*

To make navigation easier within the EA do the following:-

  1) Using Find (F3) find all instances of @@BOOKMARK
  2) On each instance press <CTRL> F2 to set the bookmark
  
Press F2 to get the next bookmark
Press <SHIFT> F2 to get the previous bookmark

*/

//+------------------------------------------------------------------+
//|                                           T101 Basket Trader.mq4 |
//|                                  Copyright  2011, John Wustrack |
//|                                        john_wustrack@hotmail.com |
//|                                                                  |
//| v1.0 - 06.09.11 - Base version                                   |
//| v1.1 - 07.09.11 - Added trailing                                 |
//|                   Fixed problem with rank values                 |
//|                   Changed min profit to breakeven                |
//|                   Added confirmation msgbox's to important       |
//|                   buttons                                        |
//|                   Added ability for pending buy/sell             |
//|                   Fixed multiple alert bug                       |
//| v1.2 - 07.09.11 - Show missing/hedged basket pairs               |
//|                   If no open basket, show general basket         |
//|                   balance                                        |
//|                   Improved robustness of opening & closing of    |
//|                   orders                                         |
//|                   Fixed currency list for basket 14              |
//|                   Remove visual nbrs for tp,sl,hedge             |
//|                   Show "PROCESSING" during open/close            |
//|                   Add ability for balance threshold protection   |
//| V1.3 - 08.09.11 - Ability to exclude pairs from basket           |
//|                   Ability to trade certain pairs opposite        |
//|                   Fixed problem with "extra" data being          |
//|                   displayed when using less than 14 pairs        |
//|                   Improved break even & lock handling            |
//|                   Moved start trailing to GVariable              |
//| V2.0 - 09.09.11 - Added Hedging                                  |
//|                   SL & TP now set in GVariable's                 |
//|                   Stop & Limit Pendings available                |
//|                   Even more checking that orders are opened      |
//|                   Redefined Nanningbob to market read            |
//|                   Ability to use Trade Logic Engines             |
//|                   Removed redundant code                         |
//|                   Orders now have the basket number              |
//|                   Different ways of showing basket balance       |
//| V2.1 - 13.09.11 - Improved pending "Limit" order processing      |
//|                   Fix for data not displayed when not using 14   |
//|                   On start-up reset the default SL,TP,Hedge      |
//|                   Add a stop & reverse button                    |
//|                   Show the last basket profit                    |
//|                   Add a suspend robot feature                    |
//|                   Controlling GVariables now have magic nbr      |
//|                   Takeprofit value can now be negative           |
//|                                                                  |
//|                                                                  |
//|                                                                  |
//+------------------------------------------------------------------+
#property copyright "Copyright  2011, John Wustrack"
#property link      "john_wustrack@hotmail.com"


//+------------------------------------------------------------------+
//| G E N E R A L   D E F I N I T I O N S                            |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
#include <WinUser32.mqh>

#define Version "v2.4 - 04 Feb 2013"
#define EAComment "Basket "
#define ObjPfx "T101 Bskt Trdr"
#define DftTrail_GVar "_Dft Start Trail at"
#define DftBrkEv_GVar "_Dft B/E+Lock at"
#define Trailing_GVar "_Trailing Value"
#define StopLoss_GVar "_Stop Loss Value"
#define TakeProfit_GVar "_Take Profit Value"
#define Hedge_GVar "_Hedge at Value"
#define MinLoss4Hedge_GVar "_Pair MinLoss for Hedge"
#define Exclusion_GVar "_Exclude "
#define LastBasketNbr_GVar "!T101 Last Basket Nbr"
#define DefaultLastBasket 999
#define ArrowPfx "T101 Trade Arrow"
#define SlippagePts 30
#define DefaultMktRead 0
#define StdCurrencyPair "EURUSD"
#define OfflineChart "CORE5"
//#define DftPendingDistance 2
#define DftExclusion 0 // 0 - Default is included, 1 - Default is excluded
#define MaxPairs 14
#define MaxRetries 60
#define NL "\n"

// Digits & Point
#define DIGITS 3
#define POINT 0.001


// Screen Definitions 
#define MktRdButX 10
#define MktRdButY 20
#define MarketReadText "Market Read - "
#define BuyButX 10
#define BuyButY 50
#define BuyText "Buy Basket"
#define SellButX 10
#define SellButY 80
#define SellText "Sell Basket"
#define CloseButX 10
#define CloseButY 50
#define CloseText "Close Basket"
#define BreakEvButX 10
#define BreakEvButY 80
#define BreakEvText "Set to Breakeven"
#define PenBuyStopButX 120
#define PenBuyStopButY 40
#define PenBuyStopText "B. Stop"
#define PenSellStopButX 120
#define PenSellStopButY 10
#define PenSellStopText "S. Stop"
#define PenBuyLimitButX 10
#define PenBuyLimitButY 40
#define PenBuyLimitText "B. Limit"
#define PenSellLimitButX 10
#define PenSellLimitButY 10
#define PenSellLimitText "S. Limit"
#define StopRevX 10
#define StopRevY 10
#define StopRevText "Stop & Reverse"

#define SmileyX 5
#define SmileyY 25
#define SmileyColor DimGray
#define SmileySize 182

#define BotCtrlX 5
#define BotCtrlY 5
#define BotStartText "START BOT"
#define BotSuspendText "SUSPEND BOT"
#define BotStartColor Green
#define BotSuspendColor DarkOrange
#define BotCtrlFont "Arial Black"
#define BotCtrlSize 12


#define LineFont "Arial"
#define LineFontSize 8

#define PandLX 10
#define PandLY 110

#define RankStartY 140
#define MisPairStartY 250
#define RankStartOffY 25
#define RankFontSize 8
int RankX[MaxPairs] = {125,10,125,10,125,10,125,10,125,10,125,10,125,10};
int RankOffY[MaxPairs] = {0,15,0,15,0,15,0,15,0,15,0,15,0,15};
#define TotBalanceX 10
#define TotBalanceY 260


#define ButtonFont "Digiface"
#define ButtonSize 18
#define DftFont "Aerial"
#define DftFontSize 14
#define DftTextColor LightGray
#define BuyColor Cyan
#define SellColor Red
#define CloseArrowColor Goldenrod
#define HedgeColor Orange
#define SLColor Red
#define TPColor Lime
#define LossColor Red
#define ProfitColor Lime
#define MarketReadColor Yellow
#define HedgeText "Hedge"
#define SLText "Stoploss"
#define TPText "TakeProfit"

#define PendingStopLine STYLE_DOT
#define PendingLimitLine STYLE_DASH

#define ProcessingFont "Arial Bold"
#define ProcessingSize 64
#define ProcessingColor Yellow
#define ProcessingX 240
#define ProcessingY 70

#define TLE_MODEOPEN 1
#define TLE_MODECLOSE 9

#define OPEN_BUY 1
#define OPEN_SELL 2
#define OPEN_BUYSTOP 3
#define OPEN_BUYLIMIT 4
#define OPEN_SELLSTOP 5
#define OPEN_SELLLIMIT 6

#define CLOSE_ORDER 9


//+------------------------------------------------------------------+
//| E N T R Y   P A R A M E T E R S                                  |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
extern string     xs_0000 = "v 2.4 - 04 Feb. 2013";
extern string     xs_0001 = "--T101 Parameters";
extern int        PairsInBasket = 5;
extern string     PairSuffix = "";
extern string     PairsToTradeOpposite = "";

extern string     xs_0002 = "--Trade/Trend Detection";
extern bool       UseMarketRead = false;
extern bool       UseTradeLogicEngine = true;
extern bool       UseTLEForOpen = true;
extern bool       UseTLEForClose = true;
extern string     TLEUsed = "!TLE_2XO";
extern bool       AutomaticTLETrading = true;

extern string     xs_0003 = "--Basket Parameters";
extern double     LotSize = 0.01;
extern double     MaxSpread         = 70;        // Maximum spread in pips, 0:Not check spread
extern int        BasketMagicNbr = 1010101;
extern int        HedgeAddMagicNbr = 90000000;

extern string     xs_0004 = "--Take Profit";
extern bool       UseTakeProfit = false;
extern bool       VisualTP = true;
extern int        DefaultTP = 2;
extern bool       ValueTP = false;
extern double     DefaultTPValue = 10;

extern string     xs_0005 = "--Stop Loss";
extern bool       UseStopLoss = true;
extern bool       VisualSL = true;
extern int        DefaultSL = 4;
extern bool       ValueSL = false;
extern double     DefaultSLValue = 12;

extern string     xs_0006 = "--Trailing & BE+Lock";
extern bool       UseTrailing = false;
extern bool       TrailFixed = false;
extern double     FixedTrailValue = 10;
extern bool       TrailPercent = false;
extern double     TrailPrcntage = 75;
extern string     xs_0006a = "--Trailing/BE+Lock Defaults";
extern double     DftTrailingAtValue = 15;
extern double     DftBreakevenAtValue = 10;
extern double     DftLockValue = 5;

extern string     xs_0007 = "--Hedging";
extern bool       AllowHedging = false;
extern bool       VisualHedge = false;
extern int        DefaultHedge = 4;
extern bool       ValueHedge = false;
extern double     DefaultHedgeValue = 100;
extern double     DefaultMinLoss4Pair = 10;

extern string     xs_0008 = "--Basket Balancing";
extern int        BasketBalanceSlowMA = 21;
extern int        BasketBalanceFastMA = 10;
extern bool       UseBsktBalLookBack = true;
extern int        TF = 10080;
extern int        BasketBalanceLB = 5;

extern string     xs_0009 = "--Other Parameters";
extern int        SleepSecs = 1;
extern int        DftPendingDistance = 2;
extern bool       ResetMktDirEveryDay = true;
extern bool       SoundOnLimitConv = true;
extern string     LimitConvSoundFile = "alert2.wav";
extern int        SizeBackGround    = 230;      // Size for background



//+------------------------------------------------------------------+
//| G L O B A L   V A R I A B L E S                                  |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
string TradePair[MaxPairs];
string PairExcluded[MaxPairs];
int TradePairOpp[MaxPairs];
int TradePairHedged[MaxPairs];
double HedgePandL[MaxPairs];
string OppTradePairs[MaxPairs];
double PairRank[MaxPairs][2];
double PairRank2[MaxPairs][2];
double PairRank3[MaxPairs][2];
double PairStrength[MaxPairs];

string MktReadText[7] = {"None","Long","Short","Range","CT Long","CT Short","n/a"};
#define MaxMktRead 5

string PendingType[7] = {"n/a","n/a","n/a","Buy Stop","Buy Limit","Sell Stop","Sell Limit"};

bool gb_Stopped;
bool gb_BotSuspended;
bool gb_OpenBasket;
bool gb_BuyAllowed;
bool gb_SellAllowed;
bool gb_PendingExists;
bool gb_ExcludedPairs;

// Hedging arrays 
#define MaxHedges 100
int HedgeXRef[100][2];  // 0-Contains the original, 1-Contains the hedge
int HedgeCount;
int HedgeMagicNbr;

double gd_Profit.Loss;
double gd_BasketBalance;
double gd_BasketBalPc;
double gi_Type;
double gd_LastBasket.PL;
double CurrentSpread;


int MarketRead;

datetime TodaysDate;
datetime LastAlert;

//+------------------------------------------------------------------+
//| E A   S T A R T I N G  /  E N D I N G                            |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| expert initialization function                                   |
//+------------------------------------------------------------------+
int init()
  {
//----
   if (PairsInBasket != 1 &&
       PairsInBasket != 2 &&
       PairsInBasket != 4 &&
       PairsInBasket != 5 &&
       PairsInBasket != 6 &&
       PairsInBasket != 7 &&
       PairsInBasket != 8 &&
       PairsInBasket != 10 &&
       PairsInBasket != 12 &&
       PairsInBasket != 14)
      {
      Alert("Invalid number of pairs in the basket, EA is stopping!");
      gb_Stopped = true;
      return(0);
      }
      
   // If Hedging is on then SL must be off
   if (AllowHedging && UseStopLoss)
      {
      Alert("Hedging & SL cannot be used together, EA is stopping!");
      gb_Stopped = true;
      return(0);
      }
  
   // If this is not the T101 offline chart then stop
  /* if (StringSubstr(Symbol(),0,4) != OfflineChart)  
      {
      Alert("This EA can ONLY be run on the T101 offline chart, EA is stopping!");
      gb_Stopped = true;
      return(0);
      }*/
      
   // If trading is not allowed then stop
   if (!IsTradeAllowed())
      {
      Alert("Trading is Allowed flag is not set, EA is stopping!");
      gb_Stopped = true;
      return(0);
      }
      
   double ld_MinLot = MarketInfo(StdCurrencyPair+PairSuffix,MODE_MINLOT);
   if (LotSize < ld_MinLot)
      {
      Alert("Lot size is less than brokers minimum lot size, setting to MinLots!");
      LotSize = ld_MinLot;
      }
      
   // If the user has selected fixed trailing and percentage - use fixed
   if (UseTrailing)
      {
      if (TrailFixed && TrailPercent) TrailPercent = false;
      if (!TrailFixed && !TrailPercent) TrailFixed = true;
      }
      
   //------------------------------------------------------
   //Background
     if(ObjectFind("Background")==-1)
     {
    ObjectCreate("Background",OBJ_LABEL,0,0,0);
    ObjectSet("Background",OBJPROP_CORNER,0);
    ObjectSet("Background",OBJPROP_BACK,FALSE);
    ObjectSet("Background",OBJPROP_YDISTANCE,13);
    ObjectSet("Background",OBJPROP_XDISTANCE,0);
    ObjectSetText("Background","g",SizeBackGround,"Webdings",DimGray);
}
//------------------------------------------------------   
  
   Get_Trade_Pairs();
   
   TodaysDate = iTime(StdCurrencyPair+PairSuffix,PERIOD_D1,0);
   
   MarketRead = DefaultMktRead;
   
   HedgeMagicNbr = BasketMagicNbr + HedgeAddMagicNbr;
   
   if (!UseTradeLogicEngine) AutomaticTLETrading = false;
   if (TLEUsed == "") UseTradeLogicEngine = false;
   
   if (!UseMarketRead) MarketRead = 6;
   
   // Delete any Pending lines that may exist
   ObjectDelete(ObjPfx+PenBuyStopText+"_VL");
   ObjectDelete(ObjPfx+PenSellStopText+"_VL");
   ObjectDelete(ObjPfx+PenBuyLimitText+"_VL");
   ObjectDelete(ObjPfx+PenSellLimitText+"_VL");
   
   
   // Create / Display the buttons
   Object_Create(ObjPfx+MarketReadText,MktRdButX,MktRdButY,MarketReadText+MktReadText[MarketRead],ButtonSize,ButtonFont,MarketReadColor);
   Setup_Buy_Sell_Buttons();
   
   // Reset Exclusion variables
   Reset_Pair_Exclusion();
   
   // Set up the global variables for breakeven & trail defaults
   GlobalVariableSet(BasketMagicNbr+DftTrail_GVar,DftTrailingAtValue);
   GlobalVariableSet(BasketMagicNbr+DftBrkEv_GVar,DftBreakevenAtValue);
   GlobalVariableSet(BasketMagicNbr+TakeProfit_GVar,DefaultTPValue);
   GlobalVariableSet(BasketMagicNbr+StopLoss_GVar,DefaultSLValue);
   GlobalVariableSet(BasketMagicNbr+Hedge_GVar,DefaultHedgeValue);
   GlobalVariableSet(BasketMagicNbr+MinLoss4Hedge_GVar,DefaultMinLoss4Pair);

   // Force the start routine to run
   start();   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| expert deinitialization function                                 |
//+------------------------------------------------------------------+
int deinit()
  {
//----
   ObjectDelete("Background");
   ObjectDelete("Spread");
   Delete_Objects();
   Comment("");
//----
   return(0);
  }



//+------------------------------------------------------------------+
//| M A I N   C O N T R O L L I N G   L O O P                        |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| expert start function                                            |
//+------------------------------------------------------------------+
int start()
  {
//----
   bool lb_ans;

   while(true)
      {
      // Exit if the EA is stopped
      if (gb_Stopped) return(0); 
      RefreshRates();
      
      Recreate_Global_Variables();
      
      // Show line texts
      Show_Line_Texts();

      // Get the pair exclusions
      Get_Pair_Exclusions();
            
      // Display the comment information
      Display_Comments();
      
      // Get the hedge trades
      Get_Basket_Hedges();
         
      gb_OpenBasket = Get_Basket_Trades();
      
      // Check if the bot is running or suspended
      if (ObjectFind(ObjPfx+"BotControl") == (-1)) Show_Bot_Control();
      
      if (ObjectGet(ObjPfx+"BotControl",OBJPROP_XDISTANCE) != BotCtrlX ||
          ObjectGet(ObjPfx+"BotControl",OBJPROP_YDISTANCE) != BotCtrlY)
         {
         if (gb_BotSuspended) 
            gb_BotSuspended = false;
         else
            gb_BotSuspended = true;
         Show_Bot_Control();  // Reset the button
         }
      
      // Show smiley and loop if suspended
      if (gb_BotSuspended)
         {
         Show_Smiley(true);
         Sleep(SleepSecs * 1000);
         continue;
         }
 
      Show_Smiley(false);
      
      // Check if any pending exists
      gb_PendingExists = false;
      if (ObjectFind(ObjPfx+PenBuyStopText+"_VL") != (-1) ||
          ObjectFind(ObjPfx+PenSellStopText+"_VL") != (-1) ||
          ObjectFind(ObjPfx+PenBuyLimitText+"_VL") != (-1) ||
          ObjectFind(ObjPfx+PenSellLimitText+"_VL") != (-1))
         gb_PendingExists = true;
            
      // If the day has changed force the user to set a direction again
      datetime ld_date = iTime(StdCurrencyPair+PairSuffix,PERIOD_D1,0);
      if (ld_date != TodaysDate && ResetMktDirEveryDay)
         {
         MarketRead=0;
         if (!UseMarketRead) MarketRead = 6;
         Setup_Buy_Sell_Buttons();
         Object_Create(ObjPfx+MarketReadText,MktRdButX,MktRdButY,MarketReadText+MktReadText[MarketRead],ButtonSize,ButtonFont,MarketReadColor);
         TodaysDate = ld_date;
         }
      
      // Delete the close basket button if no open trades
      if (!gb_OpenBasket)
         ObjectDelete(ObjPfx+CloseText);
      
      // Check for change of the direction button
      if (ObjectGet(ObjPfx+MarketReadText,OBJPROP_XDISTANCE) != MktRdButX ||
          ObjectGet(ObjPfx+MarketReadText,OBJPROP_YDISTANCE) != MktRdButY)
         {
         MarketRead++;
         if (MarketRead > MaxMktRead) MarketRead=0;
         if (!UseMarketRead) MarketRead = MaxMktRead + 1;
         // Reset the other buttons and immediately loop
         Setup_Buy_Sell_Buttons();
         Object_Create(ObjPfx+MarketReadText,MktRdButX,MktRdButY,MarketReadText+MktReadText[MarketRead],ButtonSize,ButtonFont,MarketReadColor);
         continue;
         }
         
      // If we have an open basket, perform any automatic tasks first, then manual ones
      if (gb_OpenBasket)   
         {
         // Check for hedge
         if (AllowHedging) Check_4_Hedge();
         // Check for SL
         if (UseStopLoss) Check_4_SL();
         // Check for TP
         if (UseTakeProfit) Check_4_TP();
         // Check for close
         double ld_close = 0;
         ld_close = GlobalVariableGet(BasketMagicNbr+Trailing_GVar);
         if (ld_close != 0)
            Check_4_BE(ld_close);
         // If we are using the TLE check if we should close via this   
         if (UseTradeLogicEngine && UseTLEForClose)
            Check_TLE_For_Close();
         // If we still have an open basket check for stop & reverse 
         if (gb_OpenBasket && ObjectFind(ObjPfx+StopRevText) != (-1) &&
             (ObjectGet(ObjPfx+StopRevText,OBJPROP_XDISTANCE) != StopRevX ||
              ObjectGet(ObjPfx+StopRevText,OBJPROP_YDISTANCE) != StopRevY))   
            Stop_Reverse_PP();
         // If we still have an open basket check for manual close 
         if (gb_OpenBasket && ObjectFind(ObjPfx+CloseText) != (-1) &&
             (ObjectGet(ObjPfx+CloseText,OBJPROP_XDISTANCE) != CloseButX ||
              ObjectGet(ObjPfx+CloseText,OBJPROP_YDISTANCE) != CloseButY))   
            Close_Basket_PP();
         // If we still have an open basket check for setting breakeven
         if (gb_OpenBasket && ObjectFind(ObjPfx+BreakEvText) != (-1) &&
             (ObjectGet(ObjPfx+BreakEvText,OBJPROP_XDISTANCE) != BreakEvButX ||
              ObjectGet(ObjPfx+BreakEvText,OBJPROP_YDISTANCE) != BreakEvButY))
            Set_Breakeven();   
            
         // Update the trailing amounts
         if (UseTrailing) Trail_Basket();
         
         // If we don't have an open basket anymore, reset the other buttons and immediately loop
         if (!gb_OpenBasket)
            {
            Setup_Buy_Sell_Buttons();
            continue;
            }
         // Sleep for a while, then continue 
         Sleep(SleepSecs * 1000);
         WindowRedraw();
         }// if (gb_OpenBasket)

      // Check for pre-existing pending orders 
      if (gb_PendingExists)
         {
         Open_Pendings();
         if (gb_OpenBasket) continue;
         }
         

      // Check for button press for opening new orders
      Check_4_Button_Press();
      

      if (gb_OpenBasket) continue;
      
      // If we are here, then check if there are any new trade conditions
      if (UseTradeLogicEngine && UseTLEForOpen)
         Check_TLE_For_Open();
         
      
      // Now we can sleep
      Sleep(SleepSecs * 1000);
      WindowRedraw();
      }// While(true)
   
//----
   return(0);
  }


//+------------------------------------------------------------------+
//|  E A   F U N C T I O N S                                         |
//+------------------------------------------------------------------+
// 
//
//+------------------------------------------------------------------+
//|  B A S K E T   G E T   F U N C T I O N S                         |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| get all open orders for the basket                               |
//+------------------------------------------------------------------+
bool Get_Basket_Trades()
  {
//----
   bool pb_Basket;
   color lc_Color;
   
   gd_Profit.Loss = 0;
   gi_Type = -1;
   int li_Pair;
   ArrayInitialize(PairRank,(-1));
   for (int i2=0; i2<PairsInBasket; i2++)
      PairRank[i2][0] = 0;
 
   // Get open orders   
   for (int i=0; i<=OrdersTotal(); i++)
      {
      if (OrderSelect(i-1,SELECT_BY_POS)==true)
         {
         // Not this magic number - ignore
         if (OrderMagicNumber() != BasketMagicNbr) continue;
         // Add to the Profit/Loss
         gd_Profit.Loss += OrderProfit() + OrderSwap() - OrderCommission();
         // Add to the pair rank
         li_Pair = Get_Pair_Index(OrderSymbol());
         if (li_Pair >= 0)
            {
            PairRank[li_Pair,0] += OrderProfit() + OrderSwap() - OrderCommission();
            PairRank[li_Pair,1] = li_Pair;
            }
               
         
         pb_Basket = true;
         // Save the order type, ensure we deal with opposites correctly
         if (Get_Opposite_Pair_Index(OrderSymbol()) == (-1))     
            gi_Type = OrderType();
         else 
            {
            if (OrderType() == OP_BUY) gi_Type = OP_SELL;
            if (OrderType() == OP_SELL) gi_Type = OP_BUY;
            }
         }
      }
      
   // Show the Profit / Loss
   if (pb_Basket)
      {
      double ld_hedge.total;
      for (i=0; i<PairsInBasket; i++)
          ld_hedge.total += HedgePandL[i];
      gd_Profit.Loss += ld_hedge.total;
          
      lc_Color = DftTextColor;
      
      if (gd_Profit.Loss < 0) lc_Color = LossColor;
      if (gd_Profit.Loss > 0) lc_Color = ProfitColor;
   
      string ls_Type;
      if (gi_Type == OP_BUY) ls_Type = BuyText + " P/L ";
      if (gi_Type == OP_SELL) ls_Type = SellText + " P/L ";
     
      Object_Create(ObjPfx+"PandL",PandLX,PandLY,ls_Type+DoubleToStr(gd_Profit.Loss,2)+" "+AccountCurrency(),DftFontSize,DftFont,lc_Color);
      }
   else
      ObjectDelete(ObjPfx+"PandL");
   
   double ld_hedge,ld_sl,ld_tp;
   // Delete any existing Buy/Sell buttons and create the close button if it doesnt exist
   if (pb_Basket)
      {
      if (gi_Type == OP_BUY)
         {
         ld_hedge = Close[0] - DefaultHedge;
         ld_sl = Close[0] - DefaultSL;
         ld_tp = Close[0] + DefaultTP;
         }
      if (gi_Type == OP_SELL)
         {
         ld_hedge = Close[0] + DefaultHedge;
         ld_sl = Close[0] + DefaultSL;
         ld_tp = Close[0] - DefaultTP;
         }
      
      Delete_Trading_Buttons();   
      gb_BuyAllowed = false;
      gb_SellAllowed = false;

      // Create the close, stop & reverse buttons if they dont exist
      if (ObjectFind(ObjPfx+CloseText) == (-1))
         Object_Create(ObjPfx+CloseText,CloseButX,CloseButY,CloseText,ButtonSize,ButtonFont,DftTextColor);
      if (ObjectFind(ObjPfx+StopRevText) == (-1))
         Object_Create_2(ObjPfx+StopRevText,StopRevX,StopRevY,StopRevText,ButtonSize,ButtonFont,DftTextColor);
         
      // Create the breakeven button if it doesn't exist
      if (ObjectFind(ObjPfx+BreakEvText) == (-1))
         Object_Create(ObjPfx+BreakEvText,BreakEvButX,BreakEvButY,BreakEvText,ButtonSize,ButtonFont,DftTextColor);
      
   
      // Ensure all lines that are required are still there
      if (AllowHedging)
         if (ObjectFind(ObjPfx+"HedgeLine") == (-1) && VisualHedge)
            {
            ObjectCreate(ObjPfx+"HedgeLine",OBJ_HLINE,0,0,0);
            ObjectSet(ObjPfx+"HedgeLine",OBJPROP_PRICE1,ld_hedge);
            ObjectSet(ObjPfx+"HedgeLine",OBJPROP_COLOR,HedgeColor);
            ObjectSet(ObjPfx+"HedgeLine",OBJPROP_STYLE,STYLE_DASHDOT);
            }
      if (UseStopLoss)
         if (ObjectFind(ObjPfx+"StopLoss") == (-1) && VisualSL)
            {
            ObjectCreate(ObjPfx+"StopLoss",OBJ_HLINE,0,0,0);
            ObjectSet(ObjPfx+"StopLoss",OBJPROP_PRICE1,ld_sl);
            ObjectSet(ObjPfx+"StopLoss",OBJPROP_COLOR,SLColor);
            ObjectSet(ObjPfx+"StopLoss",OBJPROP_STYLE,STYLE_DASHDOT);
            }
      if (UseTakeProfit)
         if (ObjectFind(ObjPfx+"TakeProfit") == (-1) && VisualTP)
            {
            ObjectCreate(ObjPfx+"TakeProfit",OBJ_HLINE,0,0,0);
            ObjectSet(ObjPfx+"TakeProfit",OBJPROP_PRICE1,ld_tp);
            ObjectSet(ObjPfx+"TakeProfit",OBJPROP_COLOR,TPColor);
            ObjectSet(ObjPfx+"TakeProfit",OBJPROP_STYLE,STYLE_DASHDOT);
            }
      }
      
   // If we dont have an open basket
   if (!pb_Basket)
      {
      // Delete any hedge,sl,tp lines
      ObjectDelete(ObjPfx+"HedgeLine");
      ObjectDelete(ObjPfx+"StopLoss");
      ObjectDelete(ObjPfx+"TakeProfit");
      ObjectDelete(ObjPfx+CloseText);
      ObjectDelete(ObjPfx+BreakEvText);         
      }
   
   
      
   // Show the Rankings
   Show_Pair_Rankings(pb_Basket);

//----
   return(pb_Basket);
  }
//+------------------------------------------------------------------+
//| get all hedges for the basket                                    |
//+------------------------------------------------------------------+
bool Get_Basket_Hedges()
  {
//----
   // Clear the x-ref array 
   ArrayInitialize(HedgeXRef,0);
   ArrayInitialize(TradePairHedged,0);
   ArrayInitialize(HedgePandL,0);
   HedgeCount=0;
   
   for (int i=0; i<=OrdersTotal(); i++)
      if (OrderSelect(i-1,SELECT_BY_POS)==true)
         Add_To_Hedge_Array(OrderTicket());

//----
   return(0);
  }
//+------------------------------------------------------------------+
//| add to the hedge array                                           |
//+------------------------------------------------------------------+
int Add_To_Hedge_Array(int pi_Ticket)
  {
//----
   bool lb_Ans;
   int li_Idx;
   
   // For safety reselect the order
   lb_Ans = OrderSelect(pi_Ticket,SELECT_BY_TICKET);
   if (!lb_Ans) return(0);
   
   // Ignore non-orders 
   if (OrderType() != OP_BUY && OrderType() != OP_SELL) return(0);
   
   // If not this magic number then ignore
   if (OrderMagicNumber() != HedgeMagicNbr) return(0);
   
   // Get the pair index
   li_Idx = Get_Pair_Index(OrderSymbol());
   if (li_Idx < 0) return(0);
   
   // Extract the original order from the comment 
   HedgeXRef[HedgeCount,0] = StrToInteger(OrderComment());
   HedgeXRef[HedgeCount,1] = pi_Ticket;
   TradePairHedged[li_Idx] = pi_Ticket;
   HedgePandL[li_Idx] = OrderProfit() + OrderSwap() - OrderCommission();   
   HedgeCount++;
   
//----
   return(0);
  }  



//+------------------------------------------------------------------+
//|  B A S K E T   C O N T R O L   F U N C T I O N S                 |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| check if we need to open a hedge                                 |
//+------------------------------------------------------------------+
int Check_4_Hedge()
  {
//----
   double ld_hedge;
   
   // Visually based hedging
   if (VisualHedge)
      {
      ld_hedge = ObjectGet(ObjPfx+"HedgeLine",OBJPROP_PRICE1);
      if (ld_hedge == 0) return(0);      
   
      if (gi_Type == OP_BUY) // Currently a buy open so we need to hedge with a sell
         if (Close[0] < ld_hedge)
            {
            Print("Hedging basket due to Visual Hedging");
            Hedge_Basket();
            }
         
      if (gi_Type == OP_SELL) // Currently a sell open so we need to hedge with a buy
         if (Close[0] > ld_hedge)
            {
            Print("Hedging basket due to Visual Hedging");
            Hedge_Basket();
            }
      }

   // Value based hedging
   if (ValueHedge)
      {   
      ld_hedge = GlobalVariableGet(BasketMagicNbr+Hedge_GVar);
      if (gd_Profit.Loss < (ld_hedge * (-1)) && ld_hedge != 0)
         {
         Print("Hedging basket due to Value Hedging");
         Hedge_Basket();
         }
      }
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| check if we need to close for sl                                 |
//+------------------------------------------------------------------+
int Check_4_SL()
  {
//----
   double ld_sl;
   
   // Visual SL
   if (VisualSL)
      {
      ld_sl = ObjectGet(ObjPfx+"StopLoss",OBJPROP_PRICE1);
      if (ld_sl == 0) return(0);
   
      if (gi_Type == OP_BUY) // Currently a buy open
         if (Close[0] < ld_sl)
            {
            Print("Closing basket due to Visual Stop Loss");
            Close_Basket();
            }
         
      if (gi_Type == OP_SELL) // Currently a sell open
         if (Close[0] > ld_sl)
            {
            Print("Closing basket due to Visual Stop Loss");
            Close_Basket();
            }
      }
   
   // Value SL
   if (ValueSL)
      {   
      ld_sl = GlobalVariableGet(BasketMagicNbr+StopLoss_GVar);
      if (gd_Profit.Loss < (ld_sl * (-1)) && ld_sl != 0)
         {
         Print("Closing basket due to Value Stop Loss");
         Close_Basket();
         }
      }
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| check if we need to close for tp                                 |
//+------------------------------------------------------------------+
int Check_4_TP()
  {
//----
   double ld_tp;
   
   // Visual TP
   if (VisualTP)
      {
      ld_tp = ObjectGet(ObjPfx+"TakeProfit",OBJPROP_PRICE1);
      if (ld_tp == 0) return(0);
      
   
      if (gi_Type == OP_BUY) // Currently a buy open
         if (Close[0] > ld_tp)
            {
            Print("Closing basket due to visual take profit");
            Close_Basket();
            }
         
      if (gi_Type == OP_SELL) // Currently a sell open
         if (Close[0] < ld_tp)
            {
            Print("Closing basket due to visual take profit");
            Close_Basket();
            }
      }

   // Value TP
   if (ValueTP)
      {
      ld_tp = GlobalVariableGet(BasketMagicNbr+TakeProfit_GVar);
      if ((gd_Profit.Loss > ld_tp && ld_tp > 0) ||
          (gd_Profit.Loss < ld_tp && ld_tp < 0))
         {
         Print("Closing basket due to value take profit");
         Close_Basket();
         }
      }
   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| check if we need to close for mininum profit                     |
//+------------------------------------------------------------------+
int Check_4_BE(double pd_MinP)
  {
//----
   // Value MP
   if (gd_Profit.Loss < pd_MinP && pd_MinP != 0)
      {
      Print("Closing basket due to minimum profit");
      Close_Basket();   
      }
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| trail a basket                                                   |
//+------------------------------------------------------------------+
int Trail_Basket()
  {
//----
   
   double ld_Trail = GlobalVariableGet(BasketMagicNbr+Trailing_GVar);
   double ld_BE.At = GlobalVariableGet(BasketMagicNbr+DftBrkEv_GVar);
   double ld_Trail.At = GlobalVariableGet(BasketMagicNbr+DftTrail_GVar);
   
   // Check if we need to set the breakeven + lock
   if (ld_Trail == 0)
      {
      if (gd_Profit.Loss > ld_BE.At)
         {
         ld_Trail = DftLockValue;
         GlobalVariableSet(BasketMagicNbr+Trailing_GVar,ld_Trail);
         }
      }
   
   // We only need to trail if passed the start trailing at
   if (gd_Profit.Loss < ld_Trail.At) return(0);
         
   double ld_NewTrail = 0;
   
   // Calculate the new trail value
   if (TrailFixed)
      ld_NewTrail = NormalizeDouble(gd_Profit.Loss - FixedTrailValue,2);
      
   if (TrailPercent)
      ld_NewTrail = NormalizeDouble(gd_Profit.Loss * (TrailPrcntage / 100),2);
   
   // If the new trailing value is higher than the current then update
   if (ld_NewTrail > ld_Trail)
      GlobalVariableSet(BasketMagicNbr+Trailing_GVar,ld_NewTrail);   
   
//----
   return(0);
  }




//+------------------------------------------------------------------+
//|  B A S K E T   O P E N   F U N C T I O N S                       |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| buy a basket                                                     |
//+------------------------------------------------------------------+
int Buy_Basket()
  {
//----

   // Delete the button
   Delete_Trading_Buttons();
   
   // Get the last basket number
   int BasketNbr = GlobalVariableGet(LastBasketNbr_GVar);
   BasketNbr++;
   GlobalVariableSet(LastBasketNbr_GVar,BasketNbr);
   
   Show_Processing(true);
   
   gb_BuyAllowed = false;
   
   // Reset the minimum profit
   GlobalVariableSet(BasketMagicNbr+Trailing_GVar,0);
   

   // Create the arrow 
   datetime ldt_Time = TimeCurrent();
   ObjectCreate(ArrowPfx+BuyText+ldt_Time,OBJ_ARROW,0,0,0);
   ObjectSet(ArrowPfx+BuyText+ldt_Time,OBJPROP_TIME1,Time[0]);
   ObjectSet(ArrowPfx+BuyText+ldt_Time,OBJPROP_PRICE1,Close[0]);
   ObjectSet(ArrowPfx+BuyText+ldt_Time,OBJPROP_ARROWCODE,1);
   ObjectSet(ArrowPfx+BuyText+ldt_Time,OBJPROP_COLOR,BuyColor);
   
   
   double ld_price;
   int li_ticket;
   int li_type;
   
   for (int i=0; i<PairsInBasket; i++)
      {
      if (TradePair[i] == "") break;
      if (PairExcluded[i] != "") continue;
      if (TradePairOpp[i] == 1)
         li_type = OP_SELL;
      else
         li_type = OP_BUY;

      li_ticket = Open_One_Trade(TradePair[i],li_type,LotSize,EAComment+BasketNbr,BasketMagicNbr);
      }   

   gb_OpenBasket = true;
   gi_Type = OP_BUY;
   
   Show_Processing(false);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| sell a basket                                                   |
//+------------------------------------------------------------------+
int Sell_Basket()
  {
//----
   // Delete the button
   Delete_Trading_Buttons();
   
   // Get the last basket number
   int BasketNbr = GlobalVariableGet(LastBasketNbr_GVar);
   BasketNbr++;
   GlobalVariableSet(LastBasketNbr_GVar,BasketNbr);

   Show_Processing(true);
   
   gb_SellAllowed = false;

   // Reset the minimum profit
   GlobalVariableSet(BasketMagicNbr+Trailing_GVar,0);

   // Create the arrow 
   datetime ldt_Time = TimeCurrent();
   ObjectCreate(ArrowPfx+SellText+ldt_Time,OBJ_ARROW,0,0,0);
   ObjectSet(ArrowPfx+SellText+ldt_Time,OBJPROP_TIME1,Time[0]);
   ObjectSet(ArrowPfx+SellText+ldt_Time,OBJPROP_PRICE1,Close[0]);
   ObjectSet(ArrowPfx+SellText+ldt_Time,OBJPROP_ARROWCODE,1);
   ObjectSet(ArrowPfx+SellText+ldt_Time,OBJPROP_COLOR,SellColor);

   double ld_price;
   int li_ticket;
   int li_type;
   int li_err;
   for (int i=0; i<PairsInBasket; i++)
      {
      if (TradePair[i] == "") break;
      if (PairExcluded[i] != "") continue;
      if (TradePairOpp[i] == 1)
         li_type = OP_BUY;
      else
         li_type = OP_SELL;

      li_ticket = Open_One_Trade(TradePair[i],li_type,LotSize,EAComment+BasketNbr,BasketMagicNbr);
      }   

   gb_OpenBasket = true;
   gi_Type = OP_SELL;
   
   Show_Processing(false);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| hedge a basket                                                   |
//+------------------------------------------------------------------+
int Hedge_Basket()
  {
//----   
   Show_Processing(true);
   

   // Get open orders   
   int cnt;
   int li_idx;
   double ld_profit;
   
   double ld_hedgemin = GlobalVariableGet(BasketMagicNbr+MinLoss4Hedge_GVar);
   ld_hedgemin *= (-1);
   int li_type;
   
   for (int i=0; i<=OrdersTotal(); i++)
      {
      if (OrderSelect(i-1,SELECT_BY_POS)==true)
         {
         // Not this magic number - ignore
         if (OrderMagicNumber() != BasketMagicNbr) continue;
         // Not a valid order - ignore
         if (OrderType() != OP_BUY && OrderType() != OP_SELL) continue;
         // Order already hedged
         li_idx = Get_Pair_Index(OrderSymbol());
         if (li_idx == (-1)) continue;
         if (TradePairHedged[li_idx] != 0) continue;
         // calculate the current profit
         ld_profit = OrderProfit() + OrderSwap() - OrderCommission();
            
         // if the current loss is less than hedge minimum then hedge
         if (ld_profit < ld_hedgemin)
            {
            if (OrderType() == OP_BUY) li_type = OP_SELL;
            if (OrderType() == OP_SELL) li_type = OP_BUY;
            Open_One_Trade(OrderSymbol(),li_type,OrderLots(),OrderTicket(),HedgeMagicNbr);
            }
         }
      }
      
   Show_Processing(false);
//----
   return(0);
  }

//+------------------------------------------------------------------+
//| check to see if we need to open pending baskets                  |
//+------------------------------------------------------------------+
bool Open_Pendings()
  {
//----   
   double ld_price;
   
   ld_price = ObjectGet(ObjPfx+PenBuyStopText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      {
      if (Close[0] > ld_price)
         {
         Print("Buying basket due to buy stop");
         Buy_Basket();
         return(true);
         }
      }

   ld_price = ObjectGet(ObjPfx+PenBuyLimitText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      {
      if (Close[0] < ld_price)
         {
         /*
         Print("Buying basket due to buy limit");
         Buy_Basket();  
         */
         // Convert the limit pending into a stop pending
         Convert_Limit_2_Stop(OPEN_BUYLIMIT,ld_price);
         return(true);
         }
      }
      
   ld_price = ObjectGet(ObjPfx+PenSellStopText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      {
      if (Close[0] < ld_price)
         {
         Print("Selling basket due to sell stop");
         Sell_Basket();
         return(true);
         }
      }
   ld_price = ObjectGet(ObjPfx+PenSellLimitText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      {
      if (Close[0] > ld_price)
         {
         /*
         Print("Selling basket due to sell limit");
         Sell_Basket();
         */
         // Convert the limit pending into a stop pending
         Convert_Limit_2_Stop(OPEN_SELLLIMIT,ld_price);
         return(true);
         }
      }
//----
   return(false);
  }




//+------------------------------------------------------------------+
//| B U T T O N   H A N D L I N G   R O U T I N E S                  |
//+------------------------------------------------------------------+
// @@BOOKMARK
//

//+------------------------------------------------------------------+
//| check for a button press (buy / sell)                            |
//+------------------------------------------------------------------+
int Check_4_Button_Press()
  {
//----
   // Check for manual pending buy / sell 
   if (gb_BuyAllowed && ObjectFind(ObjPfx+BuyText) != (-1) &&
       (ObjectGet(ObjPfx+BuyText,OBJPROP_XDISTANCE) != BuyButX ||
        ObjectGet(ObjPfx+BuyText,OBJPROP_YDISTANCE) != BuyButY))   
      {
      Buy_Basket_PP();
      return(0);
      }
         
   if (gb_SellAllowed && ObjectFind(ObjPfx+SellText) != (-1) &&
       (ObjectGet(ObjPfx+SellText,OBJPROP_XDISTANCE) != SellButX ||
        ObjectGet(ObjPfx+SellText,OBJPROP_YDISTANCE) != SellButY))   
      {
      Sell_Basket_PP();
      return(0);
      }
      
   // Check for manual pending buy / sell 
   if (gb_BuyAllowed && ObjectFind(ObjPfx+PenBuyStopText) != (-1) &&
       (ObjectGet(ObjPfx+PenBuyStopText,OBJPROP_XDISTANCE) != PenBuyStopButX ||
        ObjectGet(ObjPfx+PenBuyStopText,OBJPROP_YDISTANCE) != PenBuyStopButY))   
      {
      Pending_Buy_Basket_PP(OPEN_BUYSTOP);
      return(0);
      }
   if (gb_BuyAllowed && ObjectFind(ObjPfx+PenBuyLimitText) != (-1) &&
       (ObjectGet(ObjPfx+PenBuyLimitText,OBJPROP_XDISTANCE) != PenBuyLimitButX ||
        ObjectGet(ObjPfx+PenBuyLimitText,OBJPROP_YDISTANCE) != PenBuyLimitButY))   
      {
      Pending_Buy_Basket_PP(OPEN_BUYLIMIT);
      return(0);
      }
         
   if (gb_SellAllowed && ObjectFind(ObjPfx+PenSellStopText) != (-1) &&
       (ObjectGet(ObjPfx+PenSellStopText,OBJPROP_XDISTANCE) != PenSellStopButX ||
        ObjectGet(ObjPfx+PenSellStopText,OBJPROP_YDISTANCE) != PenSellStopButY))   
      {
      Pending_Sell_Basket_PP(OPEN_SELLSTOP);
      return(0);
      }
   if (gb_SellAllowed && ObjectFind(ObjPfx+PenSellLimitText) != (-1) &&
       (ObjectGet(ObjPfx+PenSellLimitText,OBJPROP_XDISTANCE) != PenSellLimitButX ||
        ObjectGet(ObjPfx+PenSellLimitText,OBJPROP_YDISTANCE) != PenSellLimitButY))   
      {
      Pending_Sell_Basket_PP(OPEN_SELLLIMIT);
      return(0);
      }

//----
   return(0);
  }
//+------------------------------------------------------------------+
//| set the breakeven                                                |
//+------------------------------------------------------------------+
int Set_Breakeven()
  {
//----
   if (gd_Profit.Loss > DftLockValue)
      GlobalVariableSet(BasketMagicNbr+Trailing_GVar,DftLockValue);
   else
      Alert("Breakeven cannot be set as the current P/L is less than the minimum!");
   Object_Create(ObjPfx+BreakEvText,BreakEvButX,BreakEvButY,BreakEvText,ButtonSize,ButtonFont,DftTextColor);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| close a basket (pre-processing)                                  |
//+------------------------------------------------------------------+
int Close_Basket_PP()
  {
//----
   int li_Ans;
   
   li_Ans = MessageBox("Are you sure you wish to close the basket?","Basket Close Confirmation",MB_YESNO|MB_ICONQUESTION);
   if (li_Ans==IDYES)
      {
      Print("Closing basket due to button press");
      Close_Basket();
      }
   else
      // Reset the button to original 
      Object_Create(ObjPfx+CloseText,CloseButX,CloseButY,CloseText,ButtonSize,ButtonFont,DftTextColor);
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| stop and reverse a basket (pre-processing)                       |
//+------------------------------------------------------------------+
int Stop_Reverse_PP()
  {
//----
   int li_Ans;
   int save_type;
   
   li_Ans = MessageBox("Are you sure you wish to stop & reverse the basket?","Stop & Reverse Confirmation",MB_YESNO|MB_ICONQUESTION);
   if (li_Ans==IDYES)
      {
      save_type = gi_Type;
      // Get rid of any lines that may cause the new order to close
      ObjectDelete(ObjPfx+"StopLoss");
      ObjectDelete(ObjPfx+"TakeProfit");
      ObjectDelete(ObjPfx+"HedgeLine");
      Print("Closing basket due to stop & reverse");
      Close_Basket();
      if (save_type == OP_BUY)
         Sell_Basket();
      if (save_type == OP_SELL)
         Buy_Basket();
      }
   else
      // Reset the button to original 
      Object_Create_2(ObjPfx+StopRevText,StopRevX,StopRevY,StopRevText,ButtonSize,ButtonFont,DftTextColor);
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| buy a basket (pre-processing)                                    |
//+------------------------------------------------------------------+
int Buy_Basket_PP()
  {
//----
   int li_Ans;
   string ls_Msg;
   
   ls_Msg = "Are you sure you wish to open a BUY basket?";
   
   if (gb_ExcludedPairs)
      ls_Msg = ls_Msg + NL + NL + "Warning: certain pairs are excluded!";
   
   li_Ans = MessageBox(ls_Msg,"BUY Basket Confirmation",MB_YESNO|MB_ICONQUESTION);
   if (li_Ans==IDYES)
      {
      Print("Buying basket due to button press");
      Buy_Basket();
      }
   else
      // Reset the button to original 
      Object_Create(ObjPfx+BuyText,BuyButX,BuyButY,BuyText,ButtonSize,ButtonFont,BuyColor);
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| sell a basket (pre-processing)                                    |
//+------------------------------------------------------------------+
int Sell_Basket_PP()
  {
//----
   int li_Ans;
   string ls_Msg;
   
   ls_Msg = "Are you sure you wish to open a SELL basket?";
   
   if (gb_ExcludedPairs)
      ls_Msg = ls_Msg + NL + NL + "Warning: certain pairs are excluded!";
   
   
   li_Ans = MessageBox(ls_Msg,"SELL Basket Confirmation",MB_YESNO|MB_ICONQUESTION);
   if (li_Ans==IDYES)
      {
      Print("Selling basket due to button press");
      Sell_Basket();
      }
   else
      // Reset the button to original 
      Object_Create(ObjPfx+SellText,SellButX,SellButY,SellText,ButtonSize,ButtonFont,SellColor);
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| pending buy basket (pre-processing)                              |
//+------------------------------------------------------------------+
int Pending_Buy_Basket_PP(int pi_Type)
  {
//----
   int li_Ans;
   string ls_Msg;

   ls_Msg = "Are you sure you wish to open a Pending " + PendingType[pi_Type];

   if (gb_ExcludedPairs)
      ls_Msg = ls_Msg + NL + NL + "Warning: certain pairs are excluded!";
   
   
   li_Ans = MessageBox(ls_Msg,"Pending BUY Confirmation",MB_YESNO|MB_ICONQUESTION);
   if (li_Ans==IDYES)
      Create_Pending_Line(pi_Type,0);

   Object_Create_2(ObjPfx+PenBuyStopText,PenBuyStopButX,PenBuyStopButY,PenBuyStopText,ButtonSize,ButtonFont,BuyColor);
   Object_Create_2(ObjPfx+PenBuyLimitText,PenBuyLimitButX,PenBuyLimitButY,PenBuyLimitText,ButtonSize,ButtonFont,BuyColor);
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| pending sell basket (pre-processing)                             |
//+------------------------------------------------------------------+
int Pending_Sell_Basket_PP(int pi_Type)
  {
//----
   int li_Ans;
   string ls_Msg;
   
   ls_Msg = "Are you sure you wish to open a Pending " + PendingType[pi_Type];
   
   if (gb_ExcludedPairs)
      ls_Msg = ls_Msg + NL + NL + "Warning: certain pairs are excluded!";
   
   li_Ans = MessageBox(ls_Msg,"Pending SELL Confirmation",MB_YESNO|MB_ICONQUESTION);
   if (li_Ans==IDYES)
      Create_Pending_Line(pi_Type,0);

   Object_Create_2(ObjPfx+PenSellStopText,PenSellStopButX,PenSellStopButY,PenSellStopText,ButtonSize,ButtonFont,SellColor);
   Object_Create_2(ObjPfx+PenSellLimitText,PenSellLimitButX,PenSellLimitButY,PenSellLimitText,ButtonSize,ButtonFont,SellColor);
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| convert a limit pending into a stop pending                      |
//+------------------------------------------------------------------+
int Convert_Limit_2_Stop(int pi_type, double pd_price)
  {
//----
   if (pi_type == OPEN_BUYLIMIT)
      {
      ObjectDelete(ObjPfx+PenBuyLimitText+"_VL");
      ObjectDelete(ObjPfx+PenBuyLimitText+"_TX");
      Create_Pending_Line(OPEN_BUYSTOP,pd_price);
      Print("Buy limit has been converted into a buy stop");
      }
      
   if (pi_type == OPEN_SELLLIMIT)
      {
      ObjectDelete(ObjPfx+PenSellLimitText+"_VL");
      ObjectDelete(ObjPfx+PenSellLimitText+"_TX");
      Create_Pending_Line(OPEN_SELLSTOP,pd_price);
      Print("Sell limit has been converted into a sell stop");
      }
  
   if (SoundOnLimitConv)
      PlaySound(LimitConvSoundFile);
   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| create a pending line                                            |
//+------------------------------------------------------------------+
int Create_Pending_Line(int pi_type, double pd_price)
  {
//----
   string ls_name;
   color lc_color;
   double ld_price;
   double li_linestyle;
   
   
   if (pi_type == OPEN_BUYSTOP)
      {
      ls_name = ObjPfx + PenBuyStopText + "_VL";
      lc_color = BuyColor;
      if (pd_price == 0)
         ld_price = Close[0] + DftPendingDistance;
      else
         ld_price = pd_price;
      li_linestyle = PendingStopLine;
      }
   if (pi_type == OPEN_BUYLIMIT)
      {
      ls_name = ObjPfx + PenBuyLimitText + "_VL";
      lc_color = BuyColor;
      if (pd_price == 0)
         ld_price = Close[0] - DftPendingDistance;
      else
         ld_price = pd_price;
      li_linestyle = PendingLimitLine;
      }

   if (pi_type == OPEN_SELLSTOP)
      {
      ls_name = ObjPfx + PenSellStopText + "_VL";
      lc_color = SellColor;
      if (pd_price == 0)
         ld_price = Close[0] - DftPendingDistance;
      else
         ld_price = pd_price;
      li_linestyle = PendingStopLine;
      }
   if (pi_type == OPEN_SELLLIMIT)
      {
      ls_name = ObjPfx + PenSellLimitText + "_VL";
      lc_color = SellColor;
      if (pd_price == 0)
         ld_price = Close[0] + DftPendingDistance;
      else
         ld_price = pd_price;
      li_linestyle = PendingLimitLine;
      }

   if (ls_name != "")
      {
      ObjectCreate(ls_name,OBJ_HLINE,0,0,0);
      ObjectSet(ls_name,OBJPROP_PRICE1,ld_price);
      ObjectSet(ls_name,OBJPROP_WIDTH,1);
      ObjectSet(ls_name,OBJPROP_COLOR,lc_color);
      ObjectSet(ls_name,OBJPROP_STYLE,li_linestyle);
      gb_PendingExists = true;
      }
   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| delete all trading buttons & lines                               |
//+------------------------------------------------------------------+
int Delete_Trading_Buttons()
  {
//----
   ObjectDelete(ObjPfx+SellText);
   ObjectDelete(ObjPfx+PenSellStopText);
   ObjectDelete(ObjPfx+PenSellLimitText);
   ObjectDelete(ObjPfx+BuyText);
   ObjectDelete(ObjPfx+PenBuyStopText);
   ObjectDelete(ObjPfx+PenBuyLimitText);

   Delete_Pending_Lines();
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| delete all pending lines                                         |
//+------------------------------------------------------------------+
int Delete_Pending_Lines()
  {
//----
   // Delete any existing line 
   ObjectDelete(ObjPfx+PenBuyStopText+"_VL");
   ObjectDelete(ObjPfx+PenSellStopText+"_VL");
   ObjectDelete(ObjPfx+PenBuyLimitText+"_VL");
   ObjectDelete(ObjPfx+PenSellLimitText+"_VL");
   ObjectDelete(ObjPfx+PenBuyStopText+"_TX");
   ObjectDelete(ObjPfx+PenSellStopText+"_TX");
   ObjectDelete(ObjPfx+PenBuyLimitText+"_TX");
   ObjectDelete(ObjPfx+PenSellLimitText+"_TX");
//----
   return(0);
  }

//+------------------------------------------------------------------+
//|  B A S K E T   C L O S E   F U N C T I O N S                     |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| close a basket                                                   |
//+------------------------------------------------------------------+
int Close_Basket()
  {
//----
   // Reset the minimum profit
   GlobalVariableSet(BasketMagicNbr+Trailing_GVar,0);
   
   // Delete the button
   ObjectDelete(ObjPfx+CloseText);
   ObjectDelete(ObjPfx+StopRevText);
   
   Show_Processing(true);
   
   // Create the arrow 
   datetime ldt_Time = TimeCurrent();
   ObjectCreate(ArrowPfx+"Close Basket"+ldt_Time,OBJ_ARROW,0,0,0);
   ObjectSet(ArrowPfx+"Close Basket"+ldt_Time,OBJPROP_TIME1,Time[0]);
   ObjectSet(ArrowPfx+"Close Basket"+ldt_Time,OBJPROP_PRICE1,Close[0]);
   ObjectSet(ArrowPfx+"Close Basket"+ldt_Time,OBJPROP_ARROWCODE,3);
   ObjectSet(ArrowPfx+"Close Basket"+ldt_Time,OBJPROP_COLOR,CloseArrowColor);

   // Get open orders   
   int cnt;
   int li_idx;
   gd_LastBasket.PL = 0;
   
   while(true)
      {
      for (int i=0; i<=OrdersTotal(); i++)
         {
         if (OrderSelect(i-1,SELECT_BY_POS)==true)
            {
            // Not this magic number - ignore
            if (OrderMagicNumber() != BasketMagicNbr) continue;
         
            li_idx = Get_Pair_Index(OrderSymbol());
            if (TradePairHedged[li_idx] == 0)
               {
               gd_LastBasket.PL += OrderProfit() + OrderSwap() - OrderCommission();
               Close_One_Trade(OrderTicket());
               }
            }
         }
      if (!Open_Trades()) break;
      cnt++;
      if (cnt > MaxRetries)
         {
         Print("Something strange happened while closing the basket, check log");
         break;
         }
      Sleep(100);
      }
   
   gb_OpenBasket = false;
   
   Show_Processing(false);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| return true if open trades                                       |
//+------------------------------------------------------------------+
bool Open_Trades()
  {
//----
   int li_idx;
   for (int i=0; i<=OrdersTotal(); i++)
      {
      if (OrderSelect(i-1,SELECT_BY_POS)==true)
         {
         // Not this magic number - ignore
         if (OrderMagicNumber() != BasketMagicNbr) continue;
         // Pair hedged - ignore
         li_idx = Get_Pair_Index(OrderSymbol());
         if (TradePairHedged[li_idx] != 0) continue;
         return(true);
         }
      }
//----
   return(false);
  }


//+------------------------------------------------------------------+
//|  O R D E R   O P E N   /   C L O S E   F U N C T I O N S         |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| open one trade                                                   |
//+------------------------------------------------------------------+
int Open_One_Trade(string ps_Symbol, int pi_Type, double pd_LotSize,
                   string ps_Comment, int pi_MagicNbr)
  {
//----
   int pi_Ticket;
   double ld_Price;
   
   int li_Cnt;
   
   while(true)
      {
      while(IsTradeContextBusy()) Sleep(100);
      double ld_Bid = MarketInfo(ps_Symbol,MODE_BID);
      double ld_Ask = MarketInfo(ps_Symbol,MODE_ASK);
      
      if (pi_Type == OP_BUY)
         ld_Price = ld_Ask;
      if (pi_Type == OP_SELL)
         ld_Price = ld_Bid;
         
      // Send the trade
      
      pi_Ticket = OrderSend(ps_Symbol,pi_Type,pd_LotSize,ld_Price,SlippagePts,0,0,ps_Comment,pi_MagicNbr,0,CLR_NONE);
      
      if (pi_Ticket > 0) break;
      
      // Otherwise we had an error, so sleep a while and try again
      li_Cnt++;
      if (li_Cnt > MaxRetries)
         {
         Print("We had a major error sending trade for ",ps_Symbol," please check logs");
         break;
         }
      Sleep(100);
      }
      
   // Ensure the trade can be found via terminal (open & history)
   li_Cnt = 0;
   int i;
   for (i=0; i<=OrdersTotal(); i++)
      {
      if (OrderSelect(i-1,SELECT_BY_POS,MODE_TRADES)==true)
         {
         if (OrderTicket() == pi_Ticket) return(0);
         li_Cnt++;
         if (li_Cnt > MaxRetries) break;
         }
      }
      
   // If we are here then the order could not be found in open orders so check history
   li_Cnt = 0;
   for (i=0; i<=OrdersHistoryTotal(); i++)
      {
      if (OrderSelect(i-1,SELECT_BY_POS,MODE_HISTORY)==true)
         {
         if (OrderTicket() == pi_Ticket) return(0);
         li_Cnt++;
         if (li_Cnt > MaxRetries) break;
         }
      }
      
   // If we are here then something seriously wrong happened
   Print("Something went seriously wrong with order send, trade not found in open or history!!");
      
      
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| close one trade                                                  |
//+------------------------------------------------------------------+
int Close_One_Trade(int pi_Ticket)
  {
//----
   int li_Cnt;
   bool lb_ans;
   
   while(true)
      {
      while(IsTradeContextBusy()) Sleep(100);
      lb_ans = OrderSelect(pi_Ticket,SELECT_BY_TICKET);
      if (!lb_ans) break;
      lb_ans = OrderClose(pi_Ticket,OrderLots(),OrderClosePrice(),SlippagePts,CLR_NONE);
      if (lb_ans) break;
      li_Cnt++;
      if (li_Cnt > MaxRetries)
         {
         Print("We had a major error closing ticket ",pi_Ticket," check the logs");
         return(0);
         }
      Sleep(100);
      }
      
//----
   return(0);
  }


//+------------------------------------------------------------------+
//| T R A D E   L O G I C   E N G I N E   R O U T I N E S            |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| check trade conditions for new entries                           |
//+------------------------------------------------------------------+
int Check_TLE_For_Open()
  {
//----
   double ld_action, ld_price;
   static double ld_prv.price;
   int NO_BSKT = -1;
   
   ld_action = iCustom(NULL,0,TLEUsed,MarketRead,TLE_MODEOPEN,NO_BSKT,0,0);
   // If the action returned is a 2-stage (ie. pending) get the price 
   if (ld_action >= 3)   
      ld_price = iCustom(NULL,0,TLEUsed,MarketRead,TLE_MODEOPEN,NO_BSKT,1,0);
      
   if (AutomaticTLETrading)
      {
      if (ld_action == OPEN_BUY)
         {
         Print("Buying basket due to TLE");
         Buy_Basket();
         }
      if (ld_action == OPEN_SELL)
         {
         Print("Selling basket due to TLE");
         Sell_Basket();
         }
      if (ld_action >= 3) // Pending
         if (ld_price != ld_prv.price)
            {
            Print("Creating a pending line due to TLE");
            Create_Pending_Line(ld_action,ld_price);
            ld_prv.price = ld_price;
            }
      }   
   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| check trade conditions for closing                               |
//+------------------------------------------------------------------+
int Check_TLE_For_Close()
  {
//----
   double ld_action;
   int BSKT;
   
   if (gi_Type == OP_BUY) BSKT = OPEN_BUY;
   if (gi_Type == OP_SELL) BSKT = OPEN_SELL;
   
   ld_action = iCustom(NULL,0,TLEUsed,MarketRead,TLE_MODECLOSE,BSKT,0,0);
   if (AutomaticTLETrading)
      if (ld_action == CLOSE_ORDER)
         {
         Print("Closing basket due to TLE");
         Close_Basket();      
         }
      
   
//----
   return(0);
  }


//+------------------------------------------------------------------+
//| S C R E E N   D I S P L A Y   R O U T I N E S                    |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| display the user comments                                        |
//+------------------------------------------------------------------+
int Display_Comments()
  {
//----
   string ls_Comment;
   int li_value;
   double ld_value;
   
/*   datetime ld_time = Time[WindowFirstVisibleBar()];
   ObjectCreate(ObjPfx+"Information",OBJ_ARROW,0,0,0,0,0);
   ObjectSet(ObjPfx+"Information",OBJPROP_PRICE1,855.220);
   ObjectSet(ObjPfx+"Information",OBJPROP_TIME1,ld_time);
   
   ObjectSet(ObjPfx+"Information",OBJPROP_ARROWCODE,255);
   ObjectSet(ObjPfx+"Information",OBJPROP_COLOR,Yellow);
*/   
   // Pairs in the basket and suffix
   ls_Comment = NL + "No. of Basket Pairs: " + PairsInBasket + " (Pair Suffix: "
                + PairSuffix + ")";
   ls_Comment = ls_Comment + NL + "Pairs to be traded oppositely: " + PairsToTradeOpposite;
   
   ls_Comment = ls_Comment + NL + "Lot Size: " + DoubleToStr(LotSize,2);
                
   // Market Read and Trade condition settings
   ls_Comment = ls_Comment + NL + NL + "Use Market Read: " + BoolToStr(UseMarketRead);
   ls_Comment = ls_Comment + NL + "Use Trade Logic Engine: " + BoolToStr(UseTradeLogicEngine);
   if (UseTradeLogicEngine)
      {
      ls_Comment = ls_Comment + NL + "......... Basket Opening: " + BoolToStr(UseTLEForOpen);
      ls_Comment = ls_Comment + "  /  Basket Closing: " + BoolToStr(UseTLEForClose);
      ls_Comment = ls_Comment + NL + "......... TLE: " + TLEUsed;
      ls_Comment = ls_Comment + NL + "......... Automatic Trading: " + BoolToStr(AutomaticTLETrading);
      }
             
   ls_Comment = ls_Comment + NL + NL + "Use Take Profit: " + BoolToStr(UseTakeProfit);

   if (UseTakeProfit)
      {
      if (VisualTP)
         {
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Visual Take Profit line shown";
         }
      if (ValueTP)
         {
         ld_value = GlobalVariableGet(BasketMagicNbr+TakeProfit_GVar);
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Value Take Profit at " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
         }
      }

   ls_Comment = ls_Comment + NL + NL + "Use Stop Loss: " + BoolToStr(UseStopLoss);
   if (UseStopLoss)
      {
      if (VisualSL)
         {
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Visual Stop Loss line shown";
         }
      if (ValueSL)
         {
         ld_value = GlobalVariableGet(BasketMagicNbr+StopLoss_GVar);
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Value Stop Loss at " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
         }
      }

   ls_Comment = ls_Comment + NL + NL + "Use Trailing: " + BoolToStr(UseTrailing);
   
   if (UseTrailing)
      {
      ld_value = GlobalVariableGet(BasketMagicNbr+DftBrkEv_GVar);
      ls_Comment = ls_Comment + NL + ".........";
      ls_Comment = ls_Comment + " Set to B/E + Lock at " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
      ld_value = GlobalVariableGet(BasketMagicNbr+DftTrail_GVar);
      ls_Comment = ls_Comment + NL + ".........";
      ls_Comment = ls_Comment + " Start trailing at " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
      
      if (TrailFixed)
         {
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Fixed trailing at " + DoubleToStr(FixedTrailValue,2) + " " + AccountCurrency();
         }
      if (TrailPercent)      
         {
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Percentage trailing at " + DoubleToStr(TrailPrcntage,0) + "%";
         }
         
      }

   // Show the current trailing / minimum profit value or last basket profit
   if (!gb_OpenBasket)
      ls_Comment = ls_Comment + NL + NL + "Last Basket Profit/Loss was " + DoubleToStr(gd_LastBasket.PL,2) + " " + AccountCurrency();
   else
      {
      ld_value = GlobalVariableGet(BasketMagicNbr+Trailing_GVar);
      ls_Comment = ls_Comment + NL + NL + "Current Minimum Profit is " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
      }

   ls_Comment = ls_Comment + NL + NL + "Hedging Allowed: " + BoolToStr(AllowHedging);
   
   if (AllowHedging)
      {
      ld_value = GlobalVariableGet(BasketMagicNbr+MinLoss4Hedge_GVar);
      ls_Comment = ls_Comment + NL + ".........";
      ls_Comment = ls_Comment + " Minimum Loss per pair to hedge: " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
      if (VisualHedge)
         {
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Visual Hedge line shown";
         }
      if (ValueHedge)
         {
         ld_value = GlobalVariableGet(BasketMagicNbr+Hedge_GVar);
         ls_Comment = ls_Comment + NL + ".........";
         ls_Comment = ls_Comment + " Value Hedge at " + DoubleToStr(ld_value,2) + " " + AccountCurrency();
         }
      }

   // Show if there is excluded pairs
   ls_Comment = ls_Comment + NL + NL + "Excluded pairs: " + BoolToStr(gb_ExcludedPairs);
   if (gb_ExcludedPairs) ls_Comment = ls_Comment + " (" + Excluded_Pair_String() + ")";
         
   ls_Comment = ls_Comment + NL + NL + "Basket Balance  (";
   if (UseBsktBalLookBack)
      ls_Comment = ls_Comment + "Lookback): ";
   else
      ls_Comment = ls_Comment + "MA X-Over): ";
   
   
   ls_Comment = ls_Comment + DoubleToStr(gd_BasketBalance,0);
   ls_Comment = ls_Comment + " / " + DoubleToStr(gd_BasketBalPc,2) + " %";
      
   Comment(ls_Comment);
//   ObjectSetText(ObjPfx+"Information",ls_Comment,20,"Arial",Blue);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Show Pair Rankings                                               |
//+------------------------------------------------------------------+
int Show_Pair_Rankings(bool pb_Basket)
  {
//----
   int i;
   int idx;
   int i2;
   string ls_pair;
   color lc_color;
   
   // If there is no basket show the general balance
   if (!pb_Basket)
      {
      for (i=0; i<PairsInBasket; i++)
         {
         ObjectDelete(ObjPfx+"MisPair"+i);
         }
      }
     
   Show_Basket_Balance(pb_Basket);

   // Otherwise show the rankings      
   if (!pb_Basket) return(0);
   
   // First of all get rid of any pairs that no longer have an order 
   ArrayCopy(PairRank3,PairRank,0,0,WHOLE_ARRAY);
   ArraySort(PairRank,0,0,MODE_DESCEND);
   ArrayInitialize(PairRank2,(-1));
   i2 = 0;
   for (i=0; i<MaxPairs; i++)
      if (PairRank[i][1] != -1)
         {
         PairRank2[i2][0] = PairRank[i][0];
         PairRank2[i2][1] = PairRank[i][1];
         i2++;
         }
   
   int OffsetY = RankStartY;
   for (i=0; i<PairsInBasket; i++)
      {
      idx = PairRank2[i][1];
      
      string ls_string = TradePair[idx] + " " + DoubleToStr(PairRank2[i][0],2) +
                         " (" + DoubleToStr(PairStrength[idx],0) + ")";
      
      // Set the color 
      lc_color = DftTextColor;
      if (PairRank2[i][0] < 0)
         lc_color = LossColor;
      if (PairRank2[i][0] > 0)
         lc_color = ProfitColor;
      if (PairRank2[i][1] != (-1))
         Object_Create(ObjPfx+"Rank"+i,RankX[i],OffsetY,ls_string,RankFontSize,DftFont,lc_color);
      else
         ObjectDelete(ObjPfx+"Rank"+i);
      OffsetY += RankOffY[i];
      }
      
   // Now show the missing pairs (and those that are hedged later)
   OffsetY = MisPairStartY;
   i2 = -1;
   for (i=0; i<PairsInBasket; i++)
      {
      if (PairRank3[i][1] == (-1))
         {
         i2++;
         ls_string = PairExcluded[i] + TradePair[i] + " (" + DoubleToStr(PairStrength[i],0) + ")";
         
         Object_Create(ObjPfx+"MisPair"+i2,RankX[i2],OffsetY,ls_string,RankFontSize,DftFont,DftTextColor);
         OffsetY += RankOffY[i2];
         }
      }
      
   // Now show the hedged pairs 
   for (i=0; i<PairsInBasket; i++)
      {
      if (TradePairHedged[i] != 0)
         {
         i2++;
         ls_string = PairExcluded[i] + TradePair[i] + " " + DoubleToStr(HedgePandL[i],2) + " (" 
                     + DoubleToStr(PairStrength[i],0) + ")";
         Object_Create(ObjPfx+"MisPair"+i2,RankX[i2],OffsetY,ls_string,RankFontSize,DftFont,HedgeColor);
         OffsetY += RankOffY[i2];
         }
      }
      
   // Get rid of any extra missing pairs that are no longer missing
   if (i2 < PairsInBasket-1)
      for (i=i2+1; i<PairsInBasket; i++)
         ObjectDelete(ObjPfx+"MisPair"+i);
         
         
   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| show the general basket balance                                  |
//+------------------------------------------------------------------+
int Show_Basket_Balance(bool pb_DontDisplay)
  {
//----
   double ld_MA.Now, ld_MA.Prv;
   double ld_MA.Fast, ld_MA.Slow;
   double ld_Balance;
   double ld_TotalBalance;
   double ld_Point;
   double ld_Buy, ld_Sell;
   double ld_Spread;
   int li_Exclusion;
   string ls_string,Spread;
   color lc_color;
   
   ArrayInitialize(PairStrength,0);
   
   int OffsetY = RankStartY;
   for (int i=0; i<PairsInBasket; i++)
      {
      // Get the MA's for now and the lookback
      if (TradePair[i] != "")
         {
         ld_Point = MarketInfo(TradePair[i],MODE_POINT);
         if (ld_Point == 0) ld_Point = 1;
         
          ld_Spread = MarketInfo(TradePair[i],MODE_SPREAD);
         
          
         ObjectCreate("Spread", OBJ_LABEL, 0, 0, 0);
         ObjectSet("Spread", OBJPROP_CORNER, 1);
         ObjectSet("Spread", OBJPROP_XDISTANCE, 10);
         ObjectSet("Spread", OBJPROP_YDISTANCE, 280);
         ObjectSet("Spread", OBJPROP_BACK, FALSE);  
         ObjectSetText("Spread" , "Total Spread  " + DoubleToStr(ld_Spread, 0), RankFontSize, DftFont, Orange);
                         
         // Determine the method of basket balancing
         if (UseBsktBalLookBack)
            {
            ld_MA.Now = iOpen(TradePair[i],TF,0);
            ld_MA.Prv = iClose(TradePair[i],TF,0);
            ld_Balance = (ld_MA.Prv - ld_MA.Now) / ld_Point / 10;
            }
       /*  else
            {
            ld_MA.Fast = iMA(TradePair[i],0,BasketBalanceFastMA,0,MODE_EMA,PRICE_CLOSE,0);
            ld_MA.Slow = iMA(TradePair[i],0,BasketBalanceSlowMA,0,MODE_EMA,PRICE_CLOSE,0);
            ld_Balance = (ld_MA.Fast - ld_MA.Slow) / ld_Point;
            }*/          
                    
         if (TradePairOpp[i] == 1) ld_Balance *= (-1);
         
         PairStrength[i] = ld_Balance;
         ld_TotalBalance += ld_Balance;
                  
         // Set the color 
         lc_color = DftTextColor;
         if (ld_Balance < 0)
            {
            lc_color = SellColor;
            ld_Sell++;
            }
         if (ld_Balance > 0)
            {
            lc_color = BuyColor;
            ld_Buy++;
            }
         
         ls_string = PairExcluded[i] + TradePair[i] + " (" + DoubleToStr(ld_Balance,0) + ")";
         if (!pb_DontDisplay)
            Object_Create(ObjPfx+"Rank"+i,RankX[i],OffsetY,ls_string,RankFontSize,DftFont,lc_color);
         OffsetY += RankOffY[i];
         }
      }

   // Now create the total balance display
   lc_color = DftTextColor;
   gd_BasketBalPc = 0;
   if (ld_TotalBalance < 0)
      {
      lc_color = SellColor;
      gd_BasketBalPc = ld_Sell / PairsInBasket * 100;
      }
   if (ld_TotalBalance > 0)
      {
      lc_color = BuyColor;
      gd_BasketBalPc = ld_Buy / PairsInBasket * 100;
      }
      
   
   ls_string = "Basket Balance " + DoubleToStr(ld_TotalBalance,0) +
                      " / " + DoubleToStr(gd_BasketBalPc,2) + " %";
   if (!pb_DontDisplay)
      Object_Create(ObjPfx+"TotalBalance",TotBalanceX,TotBalanceY,ls_string,RankFontSize,DftFont,lc_color);
      
   if (pb_DontDisplay)
      ObjectDelete(ObjPfx+"TotalBalance");
      
   gd_BasketBalance = ld_TotalBalance;         
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Show line texts                                                  |
//+------------------------------------------------------------------+
int Show_Line_Texts()
  {
//----
   double ld_price;
   
   // Pending order texts
   ld_price = ObjectGet(ObjPfx+PenSellLimitText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+PenSellLimitText+"_TX","Sell Limit",ld_price);
   else
      ObjectDelete(ObjPfx+PenSellLimitText+"_TX");
   
   ld_price = ObjectGet(ObjPfx+PenSellStopText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+PenSellStopText+"_TX","Sell Stop",ld_price);
   else
      ObjectDelete(ObjPfx+PenSellStopText+"_TX");
         
   ld_price = ObjectGet(ObjPfx+PenBuyLimitText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+PenBuyLimitText+"_TX","Buy Limit",ld_price);
   else
      ObjectDelete(ObjPfx+PenBuyLimitText+"_TX");
   
   ld_price = ObjectGet(ObjPfx+PenBuyStopText+"_VL",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+PenBuyStopText+"_TX","Buy Stop",ld_price);
   else
      ObjectDelete(ObjPfx+PenBuyStopText+"_TX");
      

   // SL,TP & Hedge
   ld_price = ObjectGet(ObjPfx+"StopLoss",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+"StopLoss_TX","Visual Stop Loss",ld_price);
   else
      ObjectDelete(ObjPfx+"StopLoss_TX");
   
   ld_price = ObjectGet(ObjPfx+"TakeProfit",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+"TakeProfit_TX","Visual Take Profit",ld_price);
   else
      ObjectDelete(ObjPfx+"TakeProfit_TX");

   ld_price = ObjectGet(ObjPfx+"HedgeLine",OBJPROP_PRICE1);
   if (ld_price != 0)
      Crt_Line_Text(ObjPfx+"HedgeLine_TX","Visual Hedge Line",ld_price);
   else
      ObjectDelete(ObjPfx+"HedgeLine_TX");
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| setup the buy sell buttons based on the market read              |
//+------------------------------------------------------------------+
int Setup_Buy_Sell_Buttons()
  {
//----
   gb_BuyAllowed = false;
   gb_SellAllowed = false;

   // If the direction is none, then delete the buy & sell
   if (MarketRead == 0)
      {
      Delete_Trading_Buttons();
/*      ObjectDelete(ObjPfx+BuyText);
      ObjectDelete(ObjPfx+SellText);
      ObjectDelete(ObjPfx+PenBuyStopText);
      ObjectDelete(ObjPfx+PenSellStopText);
      ObjectDelete(ObjPfx+PenBuyLimitText);
      ObjectDelete(ObjPfx+PenSellLimitText);  */
      return(0);
      }
      
   // If the direction is 1,3,4 or 5 we need a buy button
   if (MarketRead == 1 || MarketRead == 3 ||
       MarketRead == 4 || MarketRead == 5 || MarketRead == 6)
      {
      Object_Create(ObjPfx+BuyText,BuyButX,BuyButY,BuyText,ButtonSize,ButtonFont,BuyColor);
      gb_BuyAllowed = true;
      }
       
   // If the direction is 2,3,4 or 5 we need a sell button
   if (MarketRead == 2 || MarketRead == 3 ||
       MarketRead == 4 || MarketRead == 5 || MarketRead == 6)
      {
      Object_Create(ObjPfx+SellText,SellButX,SellButY,SellText,ButtonSize,ButtonFont,SellColor);
      gb_SellAllowed = true;
      }
      
   // If the direction is 1 we don't need a sell button
   if (MarketRead == 1)
      {
      ObjectDelete(ObjPfx+SellText);
      ObjectDelete(ObjPfx+PenSellStopText);
      ObjectDelete(ObjPfx+PenSellLimitText);
      gb_SellAllowed = false;
      }
   // If the direction is 2 we don't need a buy button
   if (MarketRead == 2)
      {
      ObjectDelete(ObjPfx+BuyText);
      ObjectDelete(ObjPfx+PenBuyStopText);
      ObjectDelete(ObjPfx+PenBuyLimitText);
      gb_BuyAllowed = false;
      }
      
   // Set up the pending buttons if they are allowed
   Create_Pending_Buttons();
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| setup the buy sell buttons based on the "direction"              |
//+------------------------------------------------------------------+
int Create_Pending_Buttons()
  {
//----
   if (gb_BuyAllowed)
      {
      Object_Create_2(ObjPfx+PenBuyStopText,PenBuyStopButX,PenBuyStopButY,PenBuyStopText,ButtonSize,ButtonFont,BuyColor);
      Object_Create_2(ObjPfx+PenBuyLimitText,PenBuyLimitButX,PenBuyLimitButY,PenBuyLimitText,ButtonSize,ButtonFont,BuyColor);
      }
   if (gb_SellAllowed)
      {
      Object_Create_2(ObjPfx+PenSellStopText,PenSellStopButX,PenSellStopButY,PenSellStopText,ButtonSize,ButtonFont,SellColor);
      Object_Create_2(ObjPfx+PenSellLimitText,PenSellLimitButX,PenSellLimitButY,PenSellLimitText,ButtonSize,ButtonFont,SellColor);
      }
//----
   return(0);
  }

//+------------------------------------------------------------------+
//| G E N E R A L   F U N C T I O N S                                |
//+------------------------------------------------------------------+
// @@BOOKMARK
//
//+------------------------------------------------------------------+
//| recreate any missing global variables                            |
//+------------------------------------------------------------------+
int Recreate_Global_Variables()
  {
//----
   if (!GlobalVariableCheck(BasketMagicNbr+Trailing_GVar))
      GlobalVariableSet(BasketMagicNbr+Trailing_GVar,0);
   if (!GlobalVariableCheck(BasketMagicNbr+TakeProfit_GVar))
      GlobalVariableSet(BasketMagicNbr+TakeProfit_GVar,DefaultTPValue);
   if (!GlobalVariableCheck(BasketMagicNbr+StopLoss_GVar))
      GlobalVariableSet(BasketMagicNbr+StopLoss_GVar,DefaultSLValue);
   if (!GlobalVariableCheck(BasketMagicNbr+Hedge_GVar))
      GlobalVariableSet(BasketMagicNbr+Hedge_GVar,DefaultHedgeValue);
   if (!GlobalVariableCheck(BasketMagicNbr+MinLoss4Hedge_GVar))
      GlobalVariableSet(BasketMagicNbr+MinLoss4Hedge_GVar,DefaultMinLoss4Pair);
   if (!GlobalVariableCheck(LastBasketNbr_GVar))
      GlobalVariableSet(LastBasketNbr_GVar,DefaultLastBasket);
         
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| reset pair exclusion                                             |
//+------------------------------------------------------------------+
int Reset_Pair_Exclusion()
  {
//----
   for (int i=0; i<PairsInBasket; i++)
      if (TradePair[i] != "")
         GlobalVariableSet(BasketMagicNbr+Exclusion_GVar+TradePair[i],DftExclusion);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| get pair exclusions                                              |
//+------------------------------------------------------------------+
int Get_Pair_Exclusions()
  {
//----
   int li_excl;
   gb_ExcludedPairs = false;
   
   for (int i=0; i<PairsInBasket; i++)
      {
      if (TradePair[i] != "")
         {
         li_excl = GlobalVariableGet(BasketMagicNbr+Exclusion_GVar+TradePair[i]);
         if (li_excl > 0)
            {
            PairExcluded[i] = "**";
            gb_ExcludedPairs = true;
            }
         else
            PairExcluded[i] = "";
         }
      }
         
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Show/Delete Processing text                                      |
//+------------------------------------------------------------------+
int Show_Processing(bool pb_show)
  {
//----
   if (pb_show)
      Object_Create(ObjPfx+"Processing",ProcessingX,ProcessingY,"PROCESSING",ProcessingSize,ProcessingFont,ProcessingColor);
   else
      ObjectDelete(ObjPfx+"Processing");
   
   WindowRedraw();   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Show Bot Control Button                                          |
//+------------------------------------------------------------------+
int Show_Bot_Control()
  {
//----
   ObjectCreate(ObjPfx+"BotControl",OBJ_LABEL,0,0,0,0,0);
   ObjectSet(ObjPfx+"BotControl",OBJPROP_CORNER,0);
   ObjectSet(ObjPfx+"BotControl",OBJPROP_XDISTANCE,BotCtrlX);
   ObjectSet(ObjPfx+"BotControl",OBJPROP_YDISTANCE,BotCtrlY);
   
   if (gb_BotSuspended)
      ObjectSetText(ObjPfx+"BotControl",BotStartText,BotCtrlSize,BotCtrlFont,BotStartColor);
   else
      ObjectSetText(ObjPfx+"BotControl",BotSuspendText,BotCtrlSize,BotCtrlFont,BotSuspendColor);
   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Show/Delete Smiley                                               |
//+------------------------------------------------------------------+
int Show_Smiley(bool pb_show)
  {
//----
   if (pb_show)
      {
      Object_Create(ObjPfx+"Smiley",SmileyX,SmileyY,CharToStr(76),SmileySize,"Wingdings",SmileyColor);
      ObjectSet(ObjPfx+"Smiley",OBJPROP_BACK,true);
      }
   else
      ObjectDelete(ObjPfx+"Smiley");
   
   WindowRedraw();   
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Show line texts                                                  |
//+------------------------------------------------------------------+
int Crt_Line_Text(string ps_obj, string ps_text, double pd_price)
  {
//----
   ObjectCreate(ps_obj,OBJ_TEXT,0,0,0);
   ObjectSet(ps_obj,OBJPROP_PRICE1,pd_price);
   ObjectSet(ps_obj,OBJPROP_TIME1,Time[0]);
   ObjectSetText(ps_obj,ps_text,LineFontSize,LineFont,DftTextColor);
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Boolean to string                                                |
//+------------------------------------------------------------------+
string BoolToStr(bool pi_bool)
  {
//----
   string ps_str;
   if (pi_bool) ps_str = "Yes";
   if (!pi_bool) ps_str = "No";
//----
   return(ps_str);
  }
//+------------------------------------------------------------------+
//| get the applicable trade pairs for the basket number             |
//+------------------------------------------------------------------+
int Get_Trade_Pairs()
  {
//----
   
   for (int i=0; i<PairsInBasket; i++)
      {
      TradePair[i] = "";
      OppTradePairs[i] = "";
      TradePairOpp[i] = 0;
      }
      
   switch (PairsInBasket)
      {
      case 1:  TradePair[0] = "GBPJPY" + PairSuffix;
               break;
      case 2:  TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURUSD" + PairSuffix;
               break;
      case 4:  TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURJPY" + PairSuffix;
               TradePair[2] = "EURUSD" + PairSuffix;
               TradePair[3] = "GBPJPY" + PairSuffix;
               break;
      case 5:  TradePair[0] = "AUDJPY" + PairSuffix;
               TradePair[1] = "NZDUSD" + PairSuffix;
               TradePair[2] = "EURJPY" + PairSuffix;
               TradePair[3] = "GBPJPY" + PairSuffix;
               TradePair[4] = "GBPUSD" + PairSuffix;
               break;         
      case 6:  TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURJPY" + PairSuffix;
               TradePair[2] = "AUDUSD" + PairSuffix;
               TradePair[3] = "EURUSD" + PairSuffix;
               TradePair[4] = "GBPJPY" + PairSuffix;
               TradePair[5] = "NZDUSD" + PairSuffix;
               break;
       case 7: TradePair[0] = "USDJPY" + PairSuffix;
               TradePair[1] = "EURJPY" + PairSuffix;
               TradePair[2] = "GBPJPY" + PairSuffix;
               TradePair[3] = "NZDJPY" + PairSuffix;
               TradePair[4] = "CHFJPY" + PairSuffix;
               TradePair[5] = "CADJPY" + PairSuffix;
               TradePair[6] = "AUDJPY" + PairSuffix;
               break;         
      case 8:  TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURJPY" + PairSuffix;
               TradePair[2] = "AUDUSD" + PairSuffix;
               TradePair[3] = "NZDJPY" + PairSuffix;
               TradePair[4] = "EURUSD" + PairSuffix;
               TradePair[5] = "GBPJPY" + PairSuffix;
               TradePair[6] = "NZDUSD" + PairSuffix;
               TradePair[7] = "AUDJPY" + PairSuffix;
               break;
      case 10: TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURGBP" + PairSuffix;
               TradePair[2] = "GBPJPY" + PairSuffix;
               TradePair[3] = "CADJPY" + PairSuffix;
               TradePair[4] = "NZDUSD" + PairSuffix;
               TradePair[5] = "EURUSD" + PairSuffix;
               TradePair[6] = "USDJPY" + PairSuffix;
               TradePair[7] = "AUDUSD" + PairSuffix;
               TradePair[8] = "NZDJPY" + PairSuffix;
               TradePair[9] = "GBPCHF" + PairSuffix;
               break;
      case 12: TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURGBP" + PairSuffix;
               TradePair[2] = "GBPJPY" + PairSuffix;
               TradePair[3] = "CADJPY" + PairSuffix;
               TradePair[4] = "NZDUSD" + PairSuffix;
               TradePair[5] = "AUDJPY" + PairSuffix;
               TradePair[6] = "EURUSD" + PairSuffix;
               TradePair[7] = "USDJPY" + PairSuffix;
               TradePair[8] = "AUDUSD" + PairSuffix;
               TradePair[9] = "NZDJPY" + PairSuffix;
               TradePair[10] = "GBPCHF" + PairSuffix;
               TradePair[11] = "CHFJPY" + PairSuffix;
               break;
      case 14: TradePair[0] = "GBPUSD" + PairSuffix;
               TradePair[1] = "EURGBP" + PairSuffix;
               TradePair[2] = "GBPJPY" + PairSuffix;
               TradePair[3] = "USDCHF" + PairSuffix;
               TradePair[4] = "NZDUSD" + PairSuffix;
               TradePair[5] = "AUDJPY" + PairSuffix;
               TradePair[6] = "EURJPY" + PairSuffix;
               TradePair[7] = "EURUSD" + PairSuffix;
               TradePair[8] = "USDJPY" + PairSuffix;
               TradePair[9] = "AUDUSD" + PairSuffix;
               TradePair[10] = "NZDJPY" + PairSuffix;
               TradePair[11] = "GBPCHF" + PairSuffix;
               TradePair[12] = "CHFJPY" + PairSuffix;
               TradePair[13] = "EURCHF" + PairSuffix;
               break;
       default: break;
       }
       
   // We now need to extract the pairs to trade opposite to the basket direction
	StrPairToStringArray(PairsToTradeOpposite, OppTradePairs, PairSuffix);
	
	// Now go through and update the array used in calculations
	for (i=0; i<PairsInBasket; i++)
	   if (Get_Opposite_Pair_Index(TradePair[i]) != (-1))
	      TradePairOpp[i] = 1;

       
//----
   return(0);
  }
//+------------------------------------------------------------------+
//| get a currency pair index                                        |
//+------------------------------------------------------------------+
int Get_Pair_Index(string ps_Pair)
  {
//----
   for (int i=0; i < PairsInBasket; i++)
      if (ps_Pair == TradePair[i]) return(i);
//----
   return(-1);
  }  
//+------------------------------------------------------------------+
//| get a opposite currency pair index                               |
//+------------------------------------------------------------------+
int Get_Opposite_Pair_Index(string ps_Pair)
  {
//----
   for (int i=0; i < PairsInBasket; i++)
      if (ps_Pair == OppTradePairs[i]) return(i);
//----
   return(-1);
  }  
//+------------------------------------------------------------------+
//| build the excluded pair string                                   |
//+------------------------------------------------------------------+
string Excluded_Pair_String()
  {
//----
   string ps_string;
   
   bool lb_ft=true;
   
   for (int i=0; i < PairsInBasket; i++)
      {
      if (PairExcluded[i] != "")
         {
         if (!lb_ft)
            ps_string = ps_string + ", " + TradePair[i];
         else
            ps_string = TradePair[i];
         lb_ft = false;
         }
      }
   
//----
   return(ps_string);
  }  
//+------------------------------------------------------------------+
//| convert string to an array                                       |
//+------------------------------------------------------------------+
void StrPairToStringArray(string str, string &a[], string p_suffix, string delim=",")
  {
//----
	int z1=-1, z2=0;
	for (int i=0; i<ArraySize(a); i++)
   	{
		z2 = StringFind(str,delim,z1+1);
		a[i] = StringSubstr(str,z1+1,z2-z1-1) + p_suffix;
		if (z2 >= StringLen(str)-1)   break;
		z1 = z2;
	   }
//----
	return(0);
  }
//+------------------------------------------------------------------+
//| create screen objects                                            |
//+------------------------------------------------------------------+
void Object_Create(string ps_name,int pi_x,int pi_y,string ps_text=" ",int pi_size=12,
                  string ps_font="Arial",color pc_colour=CLR_NONE)
  {
//----
   
   ObjectCreate(ps_name,OBJ_LABEL,0,0,0,0,0);
   ObjectSet(ps_name,OBJPROP_CORNER,1);
   ObjectSet(ps_name,OBJPROP_COLOR,pc_colour);
   ObjectSet(ps_name,OBJPROP_XDISTANCE,pi_x);
   ObjectSet(ps_name,OBJPROP_YDISTANCE,pi_y);
   
   ObjectSetText(ps_name,ps_text,pi_size,ps_font,pc_colour);

//----
   return(0);
  }
//+------------------------------------------------------------------+
//| create screen objects (bottom right)                             |
//+------------------------------------------------------------------+
void Object_Create_2(string ps_name,int pi_x,int pi_y,string ps_text=" ",int pi_size=12,
                  string ps_font="Arial",color pc_colour=CLR_NONE)
  {
//----
   
   ObjectCreate(ps_name,OBJ_LABEL,0,0,0,0,0);
   ObjectSet(ps_name,OBJPROP_CORNER,3);
   ObjectSet(ps_name,OBJPROP_COLOR,pc_colour);
   ObjectSet(ps_name,OBJPROP_XDISTANCE,pi_x);
   ObjectSet(ps_name,OBJPROP_YDISTANCE,pi_y);
   
   ObjectSetText(ps_name,ps_text,pi_size,ps_font,pc_colour);

//----
   return(0);
  }
//+------------------------------------------------------------------+
//| Delete the objects                                               |
//+------------------------------------------------------------------+
void Delete_Objects() 
  {
//----
   string ls_Obj.Name;
   for (int i=ObjectsTotal()-1; i>=0; i--)
      {
      ls_Obj.Name=ObjectName(i);
      if (StringFind(ls_Obj.Name,ObjPfx)>-1) ObjectDelete(ls_Obj.Name);
      }
   ObjectsRedraw();
//----
   return(0);
  }
//+------------------------------------------------------------------+